@@ -13,6 +13,7 @@
 # limitations under the License.
 
 import os
+import warnings
 
 import pandas as pd
 from packaging.version import Version
@@ -22,8 +23,10 @@ from mostlyai.qa.reporting import report
 from mostlyai.qa.reporting_from_statistics import report_from_statistics
 
 __all__ = ["report", "report_from_statistics", "init_logging"]
-__version__ = "1.5.2"
+__version__ = "1.5.7"
 
+warnings.filterwarnings("ignore", category=FutureWarning, module="phik")
+warnings.filterwarnings("ignore", category=UserWarning, module="huggingface_hub")
 os.environ["TOKENIZERS_PARALLELISM"] = "false"
 if Version(pd.__version__) >= Version("2.2.0"):
     pd.set_option("future.no_silent_downcasting", True)
